/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "PrintDevice.hpp"

#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>

#include "hppcl3/bitmap.hpp"

#include "version.hpp"

PrintDevice::
PrintDevice (char *pszDriverName,
             char *pszDeviceName,
             char *pszJobProperties)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << " ()" << endl;
#endif

   // Initialize class data
   outputStream_d              = &cout;
   fShouldDeleteOutputStream_d = false;
   outputStreamBuf_d           = 0;
   pszDriverName_d             = pszDriverName;
   pszDeviceName_d             = pszDeviceName;
   lCapabilities_d             = 0;
   lRasterCapabilities_d       = 0;
   lDeviceOptions_d            = 0;
   pOrientation_d              = 0;
   pszDitherID_d               = 0;
   pForm_d                     = 0;
   pTray_d                     = 0;
   pMedia_d                    = 0;
   pResolution_d               = 0;
   pCommand_d                  = 0;
   pPrintMode_d                = 0;
   pData_d                     = 0;
   pInstance_d                 = 0;
   pBlitter_d                  = 0;
   pPDL_d                      = 0;
   pszJobProperties_d          = 0;
   pszLoadedLibrary_d          = 0;
   hModLibrary_d               = 0;
   pfnOutputFunction_d         = 0;
   pMagicCookie_d              = 0;

   if (   pszJobProperties
      && *pszJobProperties
      )
   {
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
         strcpy (pszJobProperties_d, pszJobProperties);
   }
}

PrintDevice::
~PrintDevice ()
{
   if (pInstance_d)
   {
      delete pInstance_d;
      pInstance_d = 0;
   }

   if (pBlitter_d)
   {
      delete pBlitter_d;
      pBlitter_d = 0;
   }

   if (pPDL_d)
   {
      delete pPDL_d;
      pPDL_d = 0;
   }

   if (fShouldDeleteOutputStream_d)
   {
      delete outputStream_d;
      delete outputStreamBuf_d;
   }

   if (pszLoadedLibrary_d)
   {
      free (pszLoadedLibrary_d);
      pszLoadedLibrary_d = 0;
   }

   if (hModLibrary_d)
   {
      dlclose (hModLibrary_d);
      hModLibrary_d = 0;
   }

   cleanupProperties ();

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::~" << __FUNCTION__ << " ()" << endl;
#endif
}

void PrintDevice::
initialize ()
{
   initializeJobProperties ();

   if (pInstance_d)
      // Initialize the device instance
      pInstance_d->initializeInstance ();

   if (pBlitter_d)
      // Initialize the device blitter
      pBlitter_d->initializeInstance ();
}

void PrintDevice::
cleanupProperties ()
{
   if (pszJobProperties_d)
   {
      free (pszJobProperties_d);
      pszJobProperties_d = 0;
   }

   if (pOrientation_d)
   {
      delete pOrientation_d;
      pOrientation_d = 0;
   }

   if (pszDitherID_d)
   {
      free (pszDitherID_d);
      pszDitherID_d = 0;
   }

   if (pForm_d)
   {
      delete pForm_d;
      pForm_d = 0;
   }

   if (pTray_d)
   {
      delete pTray_d;
      pTray_d = 0;
   }

   if (pMedia_d)
   {
      delete pMedia_d;
      pMedia_d = 0;
   }

   if (pResolution_d)
   {
      delete pResolution_d;
      pResolution_d = 0;
   }

   if (pPrintMode_d)
   {
      delete pPrintMode_d;
      pPrintMode_d = 0;
   }

   if (pCommand_d)
   {
      delete pCommand_d;
      pCommand_d = 0;
   }

   if (pData_d)
   {
      delete pData_d;
      pData_d = 0;
   }
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

char * PrintDevice::
getVersion ()
{
   return VERSION;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

char * PrintDevice::
getDriverName ()
{
   return pszDriverName_d;
}

char * PrintDevice::
getDeviceName ()
{
   return pszDeviceName_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setDeviceInstance (DeviceInstance *pInstance)
{
   pInstance_d = pInstance;
}

DeviceInstance * PrintDevice::
getDeviceInstance ()
{
   return pInstance_d;
}

void PrintDevice::
setDeviceBlitter (DeviceBlitter *pBlitter)
{
   pBlitter_d = pBlitter;
}

void PrintDevice::
setPDL (PDL *pPDL)
{
   pPDL_d = pPDL;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setCapabilities (long lCapabilities)
{
   lCapabilities_d |= lCapabilities;
}

void PrintDevice::
setDeviceOptions (long lOptions)
{
   lDeviceOptions_d |= lOptions;
}

void PrintDevice::
setRasterCapabilities (long lRasterCapabilities)
{
   lRasterCapabilities_d |= lRasterCapabilities;
}

bool PrintDevice::
hasCapability (long lMask)
{
   if ((lCapabilities_d & lMask) != 0)
      return true;
   else
      return false;
}

bool PrintDevice::
hasRasterCapability (long lMask)
{
   if ((lRasterCapabilities_d & lMask) != 0)
      return true;
   else
      return false;
}

bool PrintDevice::
hasDeviceOption (long lMask)
{
   if ((lDeviceOptions_d & lMask) != 0)
      return true;
   else
      return false;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

int PrintDevice::
getScanlineMultiple ()
{
   int iScanLineMultiple = pResolution_d->getScanlineMultiple ();

   if (0 == iScanLineMultiple)
   {
      iScanLineMultiple = 1;

#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << " (): Error: iScanLineMultiple == 0!" << endl;
#endif
   }

   return iScanLineMultiple;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
initializeJobProperties ()
{
   // Initialize to defaults.  This is done always since you need an
   // object to create another of the same class.
   pOrientation_d = getDefaultOrientation ();
   pForm_d        = getDefaultForm ();
   pTray_d        = getDefaultTray ();
   pMedia_d       = getDefaultMedia ();
   pResolution_d  = getDefaultResolution ();
   pPrintMode_d   = getDefaultPrintMode ();

   // Allocate a copy of the dither id since it is freed
   char *pszDitherID = getDefaultDitherID ();

   if (pszDitherID)
   {
      pszDitherID_d = (char *)malloc (strlen (pszDitherID) + 1);
      if (pszDitherID_d)
         strcpy (pszDitherID_d, pszDitherID);
   }

   // These are not job properties but initialize them with the rest
   pCommand_d     = getDefaultCommands ();
   pData_d        = getDefaultData ();

   if (  pszJobProperties_d
      && *pszJobProperties_d
      )
   {
      char *pszJobProperties = pszJobProperties_d;

#ifndef RETAIL
//////if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": The job properties on input are: " << pszJobProperties << endl;
#endif

      while (  pszJobProperties
            && *pszJobProperties
            )
      {
         char *pszSep    = strpbrk (pszJobProperties, " \n\r\t");
         char *pszEquals = 0;
         char *pszLHS    = 0;
         char *pszRHS    = 0;

         if (pszSep)
         {
            // Skip white space
            while (  ' ' == pszSep[1]
                  || '\t' == pszSep[1]
                  || '\r' == pszSep[1]
                  || '\n' == pszSep[1]
                  )
               *pszSep++ = '\0';
            *pszSep = '\0';
         }

#ifndef RETAIL
/////////if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": found " << pszJobProperties << endl;
#endif

         pszEquals = strpbrk (pszJobProperties, "=");

         if (pszEquals)
         {
            *pszEquals = '\0';
            pszLHS = pszJobProperties;
            pszRHS = pszEquals + 1;

#ifndef RETAIL
////////////if (DebugOutput::shouldOutputPrintDevice ())
////////////{
////////////   cerr << "PrintDevice::" << __FUNCTION__ << ": LHS = " << pszLHS << endl;
////////////   cerr << "PrintDevice::" << __FUNCTION__ << ": RHS = " << pszRHS << endl;
////////////}
#endif

            if (0 == strcasecmp ("orientation", pszLHS))
            {
               DeviceOrientation *pDO = pOrientation_d->create (pszRHS);
               if (pDO)
               {
                  delete pOrientation_d;
                  pOrientation_d = pDO;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Orientation name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("dither", pszLHS))
            {
               if (DeviceDither::ditherNameValid (pszRHS))
               {
                  if (pszDitherID_d)
                  {
                     free (pszDitherID_d);
                     pszDitherID_d = 0;
                  }

                  pszDitherID_d = (char *)malloc (strlen (pszRHS) + 1);
                  if (pszDitherID_d)
                  {
                     strcpy (pszDitherID_d, pszRHS);
                  }
                  else
                  {
#ifndef RETAIL
                     if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": Error: out of memory!" << endl;
#endif
                  }
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Dithering name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("form", pszLHS))
            {
               DeviceForm *pDF = pForm_d->create (pszRHS);
               if (pDF)
               {
                  delete pForm_d;
                  pForm_d = pDF;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Form name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("tray", pszLHS))
            {
               DeviceTray *pDT = pTray_d->create (pszRHS);
               if (pDT)
               {
                  delete pTray_d;
                  pTray_d = pDT;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Tray name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("media", pszLHS))
            {
               DeviceMedia *pDM = pMedia_d->create (pszRHS);
               if (pDM)
               {
                  delete pMedia_d;
                  pMedia_d = pDM;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Media name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("resolution", pszLHS))
            {
               DeviceResolution *pDR = pResolution_d->create (pszRHS);
               if (pDR)
               {
                  delete pResolution_d;
                  pResolution_d = pDR;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Resolution name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("printmode", pszLHS))
            {
               DevicePrintMode *pDPM = pPrintMode_d->create (pszRHS);
               if (pDPM)
               {
                  delete pPrintMode_d;
                  pPrintMode_d = pDPM;
               }
               else
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": " << pszRHS << " is an invalid Print Mode name!" << endl;
#endif
               }
            }
            else if (0 == strcasecmp ("debugoutput", pszLHS))
            {
               if (0 == strcasecmp ("all", pszRHS))
               {
                  DebugOutput::setOutputDeviceBlitter     (true);
                  DebugOutput::setOutputDeviceDither      (true);
                  DebugOutput::setOutputDeviceInstance    (true);
                  DebugOutput::setOutputDevicePrintMode   (true);
                  DebugOutput::setOutputDeviceTester      (true);
                  DebugOutput::setOutputDeviceTester6     (true);
                  DebugOutput::setOutputGplCompression    (true);
                  DebugOutput::setOutputGplDitherInstance (true);
                  DebugOutput::setOutputOmni              (true);
                  DebugOutput::setOutputOmniInterface     (true);
                  DebugOutput::setOutputOmniProxy         (true);
                  DebugOutput::setOutputPrintDevice       (true);
                  DebugOutput::setOutputBlitter           (true);
                  DebugOutput::setOutputInstance          (true);
                  DebugOutput::setOutputOmniIJSProxy      (true);
                  DebugOutput::setOutputOmniServer        (true);
                  DebugOutput::setOutputPrinterCommand    (true);
               }
               else if (0 == strcasecmp ("DeviceBlitter", pszRHS))
               {
                  DebugOutput::setOutputDeviceBlitter (true);
               }
               else if (0 == strcasecmp ("DeviceDither", pszRHS))
               {
                  DebugOutput::setOutputDeviceDither (true);
               }
               else if (0 == strcasecmp ("DeviceInstance", pszRHS))
               {
                  DebugOutput::setOutputDeviceInstance (true);
               }
               else if (0 == strcasecmp ("DevicePrintMode", pszRHS))
               {
                  DebugOutput::setOutputDevicePrintMode (true);
               }
               else if (0 == strcasecmp ("DeviceTester", pszRHS))
               {
                  DebugOutput::setOutputDeviceTester (true);
               }
               else if (0 == strcasecmp ("DeviceTester6", pszRHS))
               {
                  DebugOutput::setOutputDeviceTester6 (true);
               }
               else if (0 == strcasecmp ("GplCompression", pszRHS))
               {
                  DebugOutput::setOutputGplCompression (true);
               }
               else if (0 == strcasecmp ("GplDitherInstance", pszRHS))
               {
                  DebugOutput::setOutputGplDitherInstance (true);
               }
               else if (0 == strcasecmp ("Omni", pszRHS))
               {
                  DebugOutput::setOutputOmni (true);
               }
               else if (0 == strcasecmp ("OmniInterface", pszRHS))
               {
                  DebugOutput::setOutputOmniInterface (true);
               }
               else if (0 == strcasecmp ("OmniProxy", pszRHS))
               {
                  DebugOutput::setOutputOmniProxy (true);
               }
               else if (0 == strcasecmp ("PrintDevice", pszRHS))
               {
                  DebugOutput::setOutputPrintDevice (true);
               }
               else if (0 == strcasecmp ("Blitter", pszRHS))
               {
                  DebugOutput::setOutputBlitter (true);
               }
               else if (0 == strcasecmp ("Instance", pszRHS))
               {
                  DebugOutput::setOutputInstance (true);
               }
               else if (0 == strcasecmp ("OmniIJSProxy", pszRHS))
               {
                  DebugOutput::setOutputOmniIJSProxy (true);
               }
               else if (0 == strcasecmp ("OmniServer", pszRHS))
               {
                  DebugOutput::setOutputOmniServer (true);
               }
               else if (0 == strcasecmp ("PrinterCommand", pszRHS))
               {
                  DebugOutput::setOutputPrinterCommand (true);
               }
               else if (0 == strcasecmp ("IJS", pszRHS))
               {
                  DebugOutput::setOutputOmniIJSProxy (true);
                  DebugOutput::setOutputOmniServer (true);
                  DebugOutput::setOutputPrinterCommand (true);
               }
            }
            else
            {
               bool rc = pInstance_d->handleDeviceJobProperty (pszLHS, pszRHS);

#ifndef RETAIL
               if (DebugOutput::shouldOutputPrintDevice ())
                  cerr << "PrintDevice::" << __FUNCTION__
                       << ": device returned "
                       << dec << rc << " from handleDeviceJobProperty ("
                       << pszLHS << ", "
                       << pszRHS << ")" << endl;
#endif
            }
         }

         if (pszSep)
            pszJobProperties = pszSep + 1;
         else
            pszJobProperties = 0;
      }

      free (pszJobProperties_d);
      pszJobProperties_d = 0;

      pForm_d->associateWith (pResolution_d);
   }
   else
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ())
         cerr << "PrintDevice::" << __FUNCTION__ << ": There are no job properties on input!" << endl;
#endif

      pForm_d->associateWith (pResolution_d);
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ())
   {
      cerr << "PrintDevice::" << __FUNCTION__ << ": The new job properties are as follows:" << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pOrientation_d->getName () << " = " << *pOrientation_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": pszDitherID_d = " << pszDitherID_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pForm_d->getName () << " = " << *pForm_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pTray_d->getName () << " = " << *pTray_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pMedia_d->getName () << " = " << *pMedia_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pResolution_d->getName () << " = " << *pResolution_d << endl;
      cerr << "PrintDevice::" << __FUNCTION__ << ": " << pPrintMode_d->getName () << " = " << *pPrintMode_d << endl;
   }
#endif
}

DeviceOrientation * PrintDevice::
getCurrentOrientation ()
{
   return pOrientation_d;
}

char * PrintDevice::
getCurrentDitherID ()
{
   return pszDitherID_d;
}

DeviceDither * PrintDevice::
getCurrentDither ()
{
   return pBlitter_d->getDitherInstance ();
}

DeviceForm * PrintDevice::
getCurrentForm ()
{
   return pForm_d;
}

DeviceTray * PrintDevice::
getCurrentTray ()
{
   return pTray_d;
}

DeviceMedia * PrintDevice::
getCurrentMedia ()
{
   return pMedia_d;
}

DeviceResolution * PrintDevice::
getCurrentResolution ()
{
   return pResolution_d;
}

DevicePrintMode * PrintDevice::
getCurrentPrintMode ()
{
   return pPrintMode_d;
}

DeviceCommand * PrintDevice::
getCommands ()
{
   return pCommand_d;
}

DeviceData * PrintDevice::
getDeviceData ()
{
   return pData_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

int PrintDevice::
getPDLLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLSubLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLSubLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLMajorRevisionLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLMajorRevisionLevel ();
   else
      return 0;
}

int PrintDevice::
getPDLMinorRevisionLevel ()
{
   if (pPDL_d)
      return pPDL_d->getPDLMinorRevisionLevel ();
   else
      return 0;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
beginJob ()
{
   return pInstance_d->beginJob ();
}

bool PrintDevice::
beginJob (char *pszJobProperties)
{
   if (pszJobProperties)
   {
      // Free the old ones
      cleanupProperties ();

      // Allocate the new one
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
      {
         strcpy (pszJobProperties_d, pszJobProperties);

         initializeJobProperties ();
      }
   }

   return pInstance_d->beginJob (true);
}

bool PrintDevice::
newFrame ()
{
   return pInstance_d->newFrame ();
}

bool PrintDevice::
newFrame (char *pszJobProperties)
{
   if (pszJobProperties)
   {
      // Free the old ones
      cleanupProperties ();

      // Allocate the new one
      pszJobProperties_d = (char *)malloc (strlen (pszJobProperties) + 1);
      if (pszJobProperties_d)
      {
         strcpy (pszJobProperties_d, pszJobProperties);

         initializeJobProperties ();
      }
   }

   return pInstance_d->newFrame (true);
}

bool PrintDevice::
endJob ()
{
   return pInstance_d->endJob ();
}

bool PrintDevice::
abortJob ()
{
   return pInstance_d->abortJob ();
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
rasterize (PBYTE        pbBits,
           PBITMAPINFO2 pbmi2,
           PRECTL       prectlPageLocation,
           BITBLT_TYPE  eType)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ())
      cerr << hex
           << "PrintDevice::rasterize (0x"
           << (int)pbBits << ", {"
           << dec << pbmi2->cx << ", "
           << pbmi2->cy << ", "
           << pbmi2->cPlanes << ", "
           << pbmi2->cBitCount << "}, "
           << "{" << prectlPageLocation->xLeft << ", " << prectlPageLocation->yBottom << ", " << prectlPageLocation->xRight << ", " << prectlPageLocation->yTop << "})"
           << endl;
#endif

   char *pszDumpEnvironmentVar = getenv ("DUMP_INCOMING_BITMAPS");
   bool  fDumpIncomingBitmaps  = false;

   if (pszDumpEnvironmentVar)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::" << __FUNCTION__ << ": (in)pszDumpEnvironmentVar = " << hex << (int)pszDumpEnvironmentVar << endl;
#endif

      if (*pszDumpEnvironmentVar)
         fDumpIncomingBitmaps = true;
   }

   if (fDumpIncomingBitmaps)
   {
      static int  iNum = 0;
      char        achName[4 + 1 + 3 + 1];
      PNEUTRALRGB pColors = 0;

      sprintf (achName, "%04dIN.bmp", iNum);

      if (8 >= pbmi2->cBitCount)
      {
         int iNumColors = 1 << pbmi2->cBitCount;

         pColors = (PNEUTRALRGB)malloc (sizeof (NEUTRALRGB) * iNumColors);
         if (pColors)
         {
            PRGB2 pRGBs = pbmi2->argbColor;

            for (int i = 0; i < iNumColors; i++)
            {
               pColors[i].bRed   = pRGBs[i].bRed;
               pColors[i].bGreen = pRGBs[i].bGreen;
               pColors[i].bBlue  = pRGBs[i].bBlue;
            }
         }
      }

      Bitmap dump (achName,
                   pbmi2->cx,
                   pbmi2->cy,
                   pbmi2->cBitCount,
                   pColors);
      dump.addScanLine (pbBits, pbmi2->cy);

      if (pColors)
      {
         free (pColors);
      }

      iNum++;
      if (999 < iNum)
         iNum = 0;
   }

   return pBlitter_d->rasterize (pbBits,
                                 pbmi2,
                                 prectlPageLocation,
                                 eType);
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
setOutputStream (ostream *osNew)
{
   outputStream_d = osNew;
}

void PrintDevice::
setOutputStream (int iFileNo)
{
   streambuf *sb = new filebuf (iFileNo);
   outputStream_d = new ostream (sb);
   fShouldDeleteOutputStream_d = true;
}

void PrintDevice::
setOutputFunction (PFNOUTPUTFUNCTION  pfnOutputFunction,
                   void              *pMagicCookie)
{
   pfnOutputFunction_d = pfnOutputFunction;
   pMagicCookie_d      = pMagicCookie;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

bool PrintDevice::
sendBinaryDataToDevice (DeviceForm *pForm)
{
   return sendBinaryDataToDevice (pForm->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceTray *pTray)
{
   return sendBinaryDataToDevice (pTray->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceMedia *pMedia)
{
   return sendBinaryDataToDevice (pMedia->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (DeviceResolution *pResolution)
{
   return sendBinaryDataToDevice (pResolution->getData ());
}

bool PrintDevice::
sendBinaryDataToDevice (BinaryData *pData)
{
   return sendBinaryDataToDevice (pData->getData (),
                                  pData->getLength ());
}

bool PrintDevice::
sendBinaryDataToDevice (PBYTE pbData, int iLength)
{
   if (pfnOutputFunction_d)
   {
      pfnOutputFunction_d (pMagicCookie_d, (unsigned char *)pbData, iLength);
   }
   else
   {
      outputStream_d->write (pbData, iLength);
      outputStream_d->flush ();
   }

   return true;
}

bool PrintDevice::
sendVPrintfToDevice (BinaryData *pData,
                     va_list     list)
{
   char    achCmd[10];
   char    achOutput[512];
   int     iBytesWritten = 0;
   bool    bRC           = false;
   char   *pszCommand    = (char *)pData->getData ();
   int     iLength       = pData->getLength ();
   int     iSrc, iDst;

   for (iSrc = 0, iDst = 0; iSrc < iLength; iSrc++)
   {
      achOutput[iDst] =pszCommand[iSrc];

      if (  '%' == pszCommand[iSrc]
         && '%' != pszCommand[iSrc+1]
         )
      {
         short sLen          = 0;
         int   iValueOnStack;

         achCmd[0] = pszCommand[iSrc];
         achCmd[1] = pszCommand[iSrc + 1];
         achCmd[2] = 0;
         iValueOnStack = va_arg (list, int);

         if ('d' == achCmd[1])
         {
            /* Special format character -- double word, little endian
            */
            achOutput[iDst]     = (iValueOnStack & 0x000000FF);
            achOutput[iDst + 1] = (iValueOnStack & 0x0000FF00) >> 8;
            achOutput[iDst + 2] = (iValueOnStack & 0x00FF0000) >> 16;
            achOutput[iDst + 3] = (iValueOnStack & 0xFF000000) >> 24;
            sLen                = 4;
         }
         else if ('D' == achCmd[1])
         {
            /* Special format character -- double word, big endian
            */
            achOutput[iDst]     = (iValueOnStack & 0xFF000000) >> 24;
            achOutput[iDst + 1] = (iValueOnStack & 0x00FF0000) >> 16;
            achOutput[iDst + 2] = (iValueOnStack & 0x0000FF00) >> 8;
            achOutput[iDst + 3] = (iValueOnStack & 0x000000FF);
            sLen                = 4;
         }
         else if ('w' == achCmd[1])
         {
            /* Special format character -- single word, little endian
            */
            achOutput[iDst]     = (iValueOnStack & 0x00FF);
            achOutput[iDst + 1] = (iValueOnStack & 0xFF00) >> 8;
            sLen                = 2;
         }
         else if ('W' == achCmd[1])
         {
            /* Special format character -- single word, big endian
            */
            achOutput[iDst]     = (iValueOnStack & 0xFF00) >> 8;
            achOutput[iDst + 1] = (iValueOnStack & 0x00FF);
            sLen                = 2;
         }
         else if (  'c' == achCmd[1]
                 || 'C' == achCmd[1]
                 )
         {
            /* Special format character -- single byte
            ** Need to special case this because if you use %c and give it
            ** a 0, then strlen will return 0 instead of 1.
            */
            achOutput[iDst] = (iValueOnStack & 0xFF);
            sLen            = 1;
         }
         else if (  'n' == achCmd[1]
                 || 'N' == achCmd[1]
                 )
         {
            /* Special format character -- ASCII number string
            */
            sprintf (achOutput + iDst, "%d", iValueOnStack);
            sLen = strlen (achOutput + iDst);
         }
         else
         {
            sprintf (achOutput + iDst,
                     achCmd,
                     iValueOnStack);

            if (iValueOnStack)
               sLen = strlen (achOutput + iDst);
            else
               sLen = 1;
         }

         iBytesWritten += sLen;
         iDst += sLen;
         iSrc++;
      }
      else if (  '%' == pszCommand[iSrc]
              && '%' == pszCommand[iSrc+1]
              )
      {
         achOutput[iDst] = '%';

         iBytesWritten++;
         iDst++;
         iSrc++;
      }
      else
      {
         iBytesWritten++;
         iDst++;
      }
   }

   if (pfnOutputFunction_d)
   {
      pfnOutputFunction_d (pMagicCookie_d, (unsigned char *)achOutput, iDst);
   }
   else
   {
      outputStream_d->write (achOutput, iDst);
      outputStream_d->flush ();
   }

   return bRC;
}

bool PrintDevice::
sendPrintfToDevice (BinaryData *pData, ...)
{
   va_list list;
   bool    rc;

   va_start (list, pData);

   rc = sendVPrintfToDevice (pData, list);

   va_end (list);

   return rc;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string * PrintDevice::
queryCurrentJobProperties ()
{
   ostrstream oss;

   oss << "orientation=" << DeviceOrientation::IDToName (pOrientation_d->getID ())
       << " dither=" << pszDitherID_d
       << " form=" << DeviceForm::IDToName (pForm_d->getID ())
       << " tray=" << DeviceTray::IDToName (pTray_d->getID ())
       << " media=" << DeviceMedia::IDToName (pMedia_d->getID ())
       << " resolution=" << DeviceResolution::IDToName (pResolution_d->getID ())
       << " printmode=" << DevicePrintMode::IDToName (pPrintMode_d->getID ());

   string *pDeviceJobProp = pInstance_d->queryDeviceJobProperties ();

   if (pDeviceJobProp)
   {
      oss << " " << *pDeviceJobProp;

      delete pDeviceJobProp;
   }

   oss << ends;

   string *pRet = new string (oss.str ());

   return pRet;
}

string * PrintDevice::
getJobProperty (char *pszKey)
{
   ostrstream oss;

   if (0 == strcasecmp ("orientation", pszKey))
   {
      oss << DeviceOrientation::IDToName (pOrientation_d->getID ());
   }
   else if (0 == strcasecmp ("dither", pszKey))
   {
      oss << pszDitherID_d;
   }
   else if (0 == strcasecmp ("form", pszKey))
   {
      oss << DeviceForm::IDToName (pForm_d->getID ());
   }
   else if (0 == strcasecmp ("tray", pszKey))
   {
      oss << DeviceTray::IDToName (pTray_d->getID ());
   }
   else if (0 == strcasecmp ("media", pszKey))
   {
      oss << DeviceMedia::IDToName (pMedia_d->getID ());
   }
   else if (0 == strcasecmp ("resolution", pszKey))
   {
      oss << DeviceResolution::IDToName (pResolution_d->getID ());
   }
   else if (0 == strcasecmp ("printmode", pszKey))
   {
      oss << DevicePrintMode::IDToName (pPrintMode_d->getID ());
   }
   else
   {
      if (pInstance_d)
      {
         string *pDeviceRet = pInstance_d->getJobProperty (pszKey);

         if (!pDeviceRet)
            return 0;

         oss << *pDeviceRet;

         delete pDeviceRet;
      }
   }

   oss << ends;

   string *pRet = new string (oss.str ());

   return pRet;
}

Enumeration * PrintDevice::
listDeviceJobPropertyKeys ()
{
   if (pInstance_d)
   {
      return pInstance_d->listDeviceJobPropertyKeys ();
   }
   else
   {
      return 0;
   }
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

DeviceBlitter * PrintDevice::
getDeviceBlitter ()
{
   return pBlitter_d;
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

void PrintDevice::
loadLibrary (char *pszLibraryName)
{
   if (pszLoadedLibrary_d)
   {
      free (pszLoadedLibrary_d);
      pszLoadedLibrary_d = 0;
   }

   if (  !pszLibraryName
      || !*pszLibraryName
      )
      return;

   hModLibrary_d = dlopen (pszLibraryName, RTLD_GLOBAL | RTLD_NOW);

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::loadLibrary: dlopen (" << pszLibraryName << ") = 0x" << hex << (int)hModLibrary_d << endl;
#endif

   if (!hModLibrary_d)
      // Error!
      return;

   pszLoadedLibrary_d = (char *)malloc (strlen (pszLibraryName) + 1);
   if (pszLoadedLibrary_d)
   {
      strcpy (pszLoadedLibrary_d, pszLibraryName);
   }
   else
   {
      dlclose (hModLibrary_d);
   }
}

void * PrintDevice::
dlsym (char *pszLibraryName,
       char *pszSymbol)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::dlsym (" << pszLibraryName << ", " << pszSymbol << ")" << endl;
#endif

   if (!pszLibraryName)
      // Nothing loaded
      return 0;

   if (0 != strcmp (pszLibraryName, pszLoadedLibrary_d))
      // Wrong library
      return 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::dlsym hModLibrary_d = 0x" << hex << (int)hModLibrary_d << endl;
#endif

   if (!hModLibrary_d)
      // Error!
      return 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputPrintDevice ()) cerr << "PrintDevice::dlsym returns " << hex << (int)::dlsym (hModLibrary_d, pszSymbol) << endl;
#endif

   return ::dlsym (hModLibrary_d, pszSymbol);
}

/*8<--------8<--------8<--------8<--------8<--------8<--------8<--------*/

string PrintDevice::
toString (ostrstream& oss)
{
   oss << hex
       << "{ lCapabilities_d = " << lCapabilities_d << ", "
       << "lRasterCapabilities_d = " << lRasterCapabilities_d << ", "
       << "lDeviceOptions_d = " << lDeviceOptions_d << ", ";

   {
      ostrstream oss2;
      if (pInstance_d)
         oss << "pInstance_d = " << pInstance_d->toString (oss2) << ", ";
      else
         oss << "pInstance_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pBlitter_d)
         oss << "pBlitter_d = " << pBlitter_d->toString (oss2) << ", ";
      else
         oss << "pBlitter_d = (null), ";
   }

   {
      ostrstream oss2;

      if (pOrientation_d)
         oss << "pOrientation_d = " << pOrientation_d->toString (oss2) << ", ";
      else
         oss << "pOrientation_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pForm_d)
         oss << "pForm_d = " << pForm_d->toString (oss2) << ", ";
      else
         oss << "pForm_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pTray_d)
         oss << "pTray_d = " << pTray_d->toString (oss2) << ", ";
      else
         oss << "pTray_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pMedia_d)
         oss << "pMedia_d = " << pMedia_d->toString (oss2) << ", ";
      else
         oss << "pMedia_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pResolution_d)
         oss << "pResolution_d = " << pResolution_d->toString (oss2) << ", ";
      else
         oss << "pResolution_d = (null), ";
   }

   if (pCommand_d)
      oss << "pCommand_d = " << pCommand_d << ", ";
   else
      oss << "pCommand_d = (null), ";

   {
      ostrstream oss2;
      if (pPrintMode_d)
         oss << "pPrintMode_d = " << pPrintMode_d->toString (oss2) << ", ";
      else
         oss << "pPrintMode_d = (null), ";
   }

   {
      ostrstream oss2;
      if (pData_d)
         oss << "pData_d = " << pData_d->toString (oss2);
      else
         oss << "pData_d = (null)";
   }

   oss << " }"
       << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const PrintDevice& const_self)
{
   PrintDevice& self = const_cast<PrintDevice&>(const_self);
   ostrstream   oss;

   os << self.toString (oss);

   return os;
}
